def subtract(minuend, subtrahend):
    """Subtract the second variable from the first.

    Parameters
    ----------
    minuend : int or float
        The number to subtract from.
    subtrahend : int or float
        The number to subtract.
    """
    return minuend - subtrahend
